"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const chai = require("chai");
const cancellation_1 = require("../cancellation");
const chaiAsPromised = require("chai-as-promised");
const sinon = require("sinon");
chai.use(chaiAsPromised);
const assert = chai.assert;
describe('cancellation', () => {
    describe('cancellableMemoize()', () => {
        it('should memoize a function by the first argument', () => __awaiter(this, void 0, void 0, function* () {
            const toBeMemoized = sinon.spy((arg1, arg2, token = cancellation_1.CancellationToken.None) => {
                return Promise.resolve(Math.random());
            });
            const memoized = cancellation_1.cancellableMemoize(toBeMemoized);
            const a = yield memoized(123, 456);
            const b = yield memoized(123, 456);
            sinon.assert.calledOnce(toBeMemoized);
            sinon.assert.calledWith(toBeMemoized, 123, 456, sinon.match.object);
            assert.equal(a, b);
        }));
        it('should memoize a function without parameters', () => __awaiter(this, void 0, void 0, function* () {
            const toBeMemoized = sinon.spy((token = cancellation_1.CancellationToken.None) => {
                return Promise.resolve(Math.random());
            });
            const memoized = cancellation_1.cancellableMemoize(toBeMemoized);
            const a = yield memoized();
            const b = yield memoized();
            assert.equal(a, b);
            sinon.assert.calledOnce(toBeMemoized);
            sinon.assert.calledWith(toBeMemoized, sinon.match.object);
        }));
        it('should not cancel the operation if there are still consumers', () => __awaiter(this, void 0, void 0, function* () {
            const toBeMemoized = sinon.spy((arg, token = cancellation_1.CancellationToken.None) => {
                return new Promise((resolve, reject) => {
                    token.onCancellationRequested(() => {
                        reject(new cancellation_1.CancelledError());
                    });
                    setTimeout(() => resolve(123), 500);
                });
            });
            const memoized = cancellation_1.cancellableMemoize(toBeMemoized);
            const source1 = new cancellation_1.CancellationTokenSource();
            const source2 = new cancellation_1.CancellationTokenSource();
            const promise1 = memoized(123, source1.token);
            const promise2 = memoized(123, source2.token);
            source1.cancel();
            assert.equal(yield promise1, 123);
            assert.equal(yield promise2, 123);
            sinon.assert.calledOnce(toBeMemoized);
            sinon.assert.calledWith(toBeMemoized, 123, sinon.match.object);
        }));
        it('should cancel the operation if all consumers requested cancellation', () => __awaiter(this, void 0, void 0, function* () {
            const toBeMemoized = sinon.spy((arg, token) => {
                return new Promise((resolve, reject) => {
                    token.onCancellationRequested(() => {
                        reject(new cancellation_1.CancelledError());
                    });
                    setTimeout(() => resolve(123), 500);
                });
            });
            const memoized = cancellation_1.cancellableMemoize(toBeMemoized);
            const source1 = new cancellation_1.CancellationTokenSource();
            const source2 = new cancellation_1.CancellationTokenSource();
            const promise1 = memoized(123, source1.token);
            const promise2 = memoized(123, source2.token);
            source1.cancel();
            source2.cancel();
            sinon.assert.calledOnce(toBeMemoized);
            sinon.assert.calledWith(toBeMemoized, 123, sinon.match.object);
            yield assert.isRejected(promise1, /cancel/i);
            yield assert.isRejected(promise2, /cancel/i);
        }));
    });
});
//# sourceMappingURL=cancellation.test.js.map