"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("@reactivex/rxjs");
const iterare_1 = require("iterare");
const lodash_1 = require("lodash");
const opentracing_1 = require("opentracing");
const path_ = require("path");
const ts = require("typescript");
const url = require("url");
const vscode_languageserver_1 = require("vscode-languageserver");
const ast_1 = require("./ast");
const fs_1 = require("./fs");
const logging_1 = require("./logging");
const memfs_1 = require("./memfs");
const pm = require("./project-manager");
const util = require("./util");
/**
 * Handles incoming requests and return responses. There is a one-to-one-to-one
 * correspondence between TCP connection, TypeScriptService instance, and
 * language workspace. TypeScriptService caches data from the compiler across
 * requests. The lifetime of the TypeScriptService instance is tied to the
 * lifetime of the TCP connection, so its caches are deleted after the
 * connection is torn down.
 *
 * Methods are camelCase versions of the LSP spec methods and dynamically
 * dispatched. Methods not to be exposed over JSON RPC are prefixed with an
 * underscore.
 */
class TypeScriptService {
    constructor(client, options = {}) {
        this.client = client;
        this.options = options;
        this.logger = new logging_1.LSPLogger(client);
    }
    /**
     * The initialize request is sent as the first request from the client to the server. If the
     * server receives request or notification before the `initialize` request it should act as
     * follows:
     *
     * - for a request the respond should be errored with `code: -32002`. The message can be picked by
     * the server.
     * - notifications should be dropped, except for the exit notification. This will allow the exit a
     * server without an initialize request.
     *
     * Until the server has responded to the `initialize` request with an `InitializeResult` the
     * client must not sent any additional requests or notifications to the server.
     *
     * During the `initialize` request the server is allowed to sent the notifications
     * `window/showMessage`, `window/logMessage` and `telemetry/event` as well as the
     * `window/showMessageRequest` request to the client.
     */
    initialize(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            if (params.rootUri || params.rootPath) {
                this.root = params.rootPath || util.uri2path(params.rootUri);
                this.rootUri = params.rootUri || util.path2uri('', params.rootPath);
                this._initializeFileSystems(!this.options.strict && !(params.capabilities.xcontentProvider && params.capabilities.xfilesProvider));
                this.updater = new fs_1.FileSystemUpdater(this.fileSystem, this.inMemoryFileSystem);
                this.projectManager = new pm.ProjectManager(this.root, this.inMemoryFileSystem, this.updater, !!this.options.strict, this.traceModuleResolution, this.logger);
                // Pre-fetch files in the background
                this.projectManager.ensureOwnFiles(span).catch(err => {
                    this.logger.error('Background fetching failed ', err);
                });
            }
            return {
                capabilities: {
                    // Tell the client that the server works in FULL text document sync mode
                    textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Full,
                    hoverProvider: true,
                    signatureHelpProvider: {
                        triggerCharacters: ['(', ',']
                    },
                    definitionProvider: true,
                    referencesProvider: true,
                    documentSymbolProvider: true,
                    workspaceSymbolProvider: true,
                    xworkspaceReferencesProvider: true,
                    xdefinitionProvider: true,
                    xdependenciesProvider: true,
                    completionProvider: {
                        resolveProvider: false,
                        triggerCharacters: ['.']
                    },
                    xpackagesProvider: true
                }
            };
        });
    }
    /**
     * Initializes the remote file system and in-memory file system.
     * Can be overridden
     *
     * @param accessDisk Whether the language server is allowed to access the local file system
     */
    _initializeFileSystems(accessDisk) {
        this.fileSystem = accessDisk ? new fs_1.LocalFileSystem(util.uri2path(this.root)) : new fs_1.RemoteFileSystem(this.client);
        this.inMemoryFileSystem = new memfs_1.InMemoryFileSystem(this.root);
    }
    /**
     * The shutdown request is sent from the client to the server. It asks the server to shut down,
     * but to not exit (otherwise the response might not be delivered correctly to the client).
     * There is a separate exit notification that asks the server to exit.
     */
    shutdown(params = {}, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            this.projectManager.dispose();
            return null;
        });
    }
    /**
     * The goto definition request is sent from the client to the server to resolve the definition
     * location of a symbol at a given text document position.
     */
    textDocumentDefinition(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            const line = params.position.line;
            const column = params.position.character;
            // Fetch files needed to resolve definition
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                return [];
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, line, column);
            const defs = configuration.getService().getDefinitionAtPosition(fileName, offset);
            const ret = [];
            if (defs) {
                for (const def of defs) {
                    const sourceFile = this._getSourceFile(configuration, def.fileName, span);
                    if (!sourceFile) {
                        throw new Error('expected source file "' + def.fileName + '" to exist in configuration');
                    }
                    const start = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start);
                    const end = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start + def.textSpan.length);
                    ret.push(vscode_languageserver_1.Location.create(this._defUri(def.fileName), {
                        start,
                        end
                    }));
                }
            }
            return ret;
        });
    }
    /**
     * This method is the same as textDocument/definition, except that:
     *
     * - The method returns metadata about the definition (the same metadata that
     * workspace/xreferences searches for).
     * - The concrete location to the definition (location field)
     * is optional. This is useful because the language server might not be able to resolve a goto
     * definition request to a concrete location (e.g. due to lack of dependencies) but still may
     * know some information about it.
     */
    textDocumentXdefinition(params, span = new opentracing_1.Span()) {
        // Ensure files needed to resolve SymbolLocationInformation are fetched
        return this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span)
            .toArray()
            .mergeMap(() => {
            // Convert URI to file path
            const fileName = util.uri2path(params.textDocument.uri);
            // Get closest tsconfig configuration
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Unknown text document ${params.textDocument.uri}`);
            }
            // Convert line/character to offset
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            // Query TypeScript for references
            return rxjs_1.Observable.from(configuration.getService().getDefinitionAtPosition(fileName, offset) || [])
                .map(def => {
                const sourceFile = this._getSourceFile(configuration, def.fileName, span);
                if (!sourceFile) {
                    throw new Error(`Expected source file ${def.fileName} to exist in configuration`);
                }
                // Convert offset to line/character
                const start = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start);
                const end = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start + def.textSpan.length);
                return {
                    symbol: util.defInfoToSymbolDescriptor(def),
                    location: {
                        uri: this._defUri(def.fileName),
                        range: {
                            start,
                            end
                        }
                    }
                };
            });
        })
            .toArray();
    }
    /**
     * The hover request is sent from the client to the server to request hover information at a
     * given text document position.
     */
    textDocumentHover(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve hover are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Unknown text document ${params.textDocument.uri}`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const info = configuration.getService().getQuickInfoAtPosition(fileName, offset);
            if (!info) {
                return { contents: [] };
            }
            const contents = [{
                    language: 'typescript',
                    value: ts.displayPartsToString(info.displayParts)
                }];
            const documentation = ts.displayPartsToString(info.documentation);
            if (documentation) {
                contents.push(documentation);
            }
            const start = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start);
            const end = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start + info.textSpan.length);
            return { contents, range: vscode_languageserver_1.Range.create(start.line, start.character, end.line, end.character) };
        });
    }
    /**
     * The references request is sent from the client to the server to resolve project-wide
     * references for the symbol denoted by the given text document position.
     *
     * Returns all references to the symbol at the position in the own workspace, including references inside node_modules.
     */
    textDocumentReferences(params, span = new opentracing_1.Span()) {
        // Ensure all files were fetched to collect all references
        return rxjs_1.Observable.from(this.projectManager.ensureOwnFiles(span))
            .mergeMap(() => {
            // Convert URI to file path because TypeScript doesn't work with URIs
            const fileName = util.uri2path(params.textDocument.uri);
            // Get tsconfig configuration for requested file
            const configuration = this.projectManager.getConfiguration(fileName);
            // Ensure all files have been added
            configuration.ensureAllFiles(span);
            const program = configuration.getProgram();
            if (!program) {
                return [];
            }
            // Get SourceFile object for requested file
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Source file ${fileName} does not exist`);
            }
            // Convert line/character to offset
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            // Request references at position from TypeScript
            // Despite the signature, getReferencesAtPosition() can return undefined
            return rxjs_1.Observable.from(configuration.getService().getReferencesAtPosition(fileName, offset) || [])
                .filter(reference => 
            // Filter declaration if not requested
            (!reference.isDefinition || (params.context && params.context.includeDeclaration))
                && !reference.fileName.includes('/node_modules/'))
                .map(reference => {
                const sourceFile = program.getSourceFile(reference.fileName);
                if (!sourceFile) {
                    throw new Error(`Source file ${reference.fileName} does not exist`);
                }
                // Convert offset to line/character position
                const start = ts.getLineAndCharacterOfPosition(sourceFile, reference.textSpan.start);
                const end = ts.getLineAndCharacterOfPosition(sourceFile, reference.textSpan.start + reference.textSpan.length);
                return {
                    uri: util.path2uri(this.root, reference.fileName),
                    range: {
                        start,
                        end
                    }
                };
            });
        })
            .toArray();
    }
    /**
     * The workspace symbol request is sent from the client to the server to list project-wide
     * symbols matching the query string. The text document parameter specifies the active document
     * at time of the query. This can be used to rank or limit results.
     */
    workspaceSymbol(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Always return max. 50 results
            // TODO stream 50 results, then re-query and stream the rest
            const limit = Math.min(params.limit || Infinity, 50);
            const query = params.query;
            const symbolQuery = params.symbol ? Object.assign({}, params.symbol) : undefined;
            if (symbolQuery && symbolQuery.package) {
                symbolQuery.package = { name: symbolQuery.package.name };
            }
            if (symbolQuery) {
                try {
                    const dtRes = yield this._workspaceSymbolDefinitelyTyped(Object.assign({}, params, { limit }));
                    if (dtRes) {
                        return dtRes;
                    }
                }
                catch (err) {
                    // Ignore
                }
                if (!symbolQuery.containerKind) {
                    // symbolQuery.containerKind is sometimes empty when symbol.containerKind = 'module'
                    symbolQuery.containerKind = undefined;
                }
            }
            // A workspace/symol request searches all symbols in own code, but not in dependencies
            yield this.projectManager.ensureOwnFiles(span);
            // Cache result for empty query
            if (!query && !symbolQuery && this.emptyQueryWorkspaceSymbols) {
                return this.emptyQueryWorkspaceSymbols;
            }
            // Find configurations to search
            let configs;
            if (symbolQuery && symbolQuery.package && symbolQuery.package.name) {
                // If PackageDescriptor is given, only search project with the matching package name
                configs = iterare_1.default(this.projectManager.configurations())
                    .filter(config => config.getPackageName() === symbolQuery.package.name);
            }
            else {
                const rootConfig = this.projectManager.getConfiguration('');
                if (rootConfig) {
                    // Use root configuration because it includes all files
                    configs = [rootConfig];
                }
                else {
                    // Use all configurations
                    configs = this.projectManager.configurations();
                }
            }
            const symbols = iterare_1.default(configs)
                .map(config => this._collectWorkspaceSymbols(config, query || symbolQuery, limit))
                .flatten()
                .take(limit)
                .toArray();
            // Save empty query result
            if (!query && !symbolQuery) {
                this.emptyQueryWorkspaceSymbols = symbols;
            }
            return symbols;
        });
    }
    _workspaceSymbolDefinitelyTyped(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const rootUriParts = url.parse(this.rootUri);
            if (!rootUriParts.pathname) {
                return null;
            }
            const packageJsonUri = url.format(Object.assign({}, rootUriParts, { pathname: path_.posix.join(rootUriParts.pathname, 'package.json') }));
            yield this.updater.ensure(packageJsonUri);
            const rootConfig = JSON.parse(this.inMemoryFileSystem.getContent(packageJsonUri));
            if (rootConfig.name !== 'definitely-typed') {
                return null;
            }
            if (!params.symbol || !params.symbol.package) {
                return null;
            }
            const pkg = params.symbol.package;
            if (!pkg.name || !pkg.name.startsWith('@types/')) {
                return null;
            }
            const relPkgRoot = pkg.name.slice('@types/'.length);
            yield this.projectManager.ensureModuleStructure();
            const symbolQuery = params.symbol ? Object.assign({}, params.symbol) : undefined;
            if (symbolQuery) {
                symbolQuery.package = undefined;
                if (!symbolQuery.containerKind) {
                    symbolQuery.containerKind = undefined; // symQuery.containerKind is sometimes empty when symbol.containerKind = 'module'
                }
            }
            const config = this.projectManager.getConfiguration(relPkgRoot);
            return Array.from(this._collectWorkspaceSymbols(config, params.query || symbolQuery, params.limit));
        });
    }
    /**
     * The document symbol request is sent from the client to the server to list all symbols found
     * in a given text document.
     */
    textDocumentDocumentSymbol(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve symbols are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const config = this.projectManager.getConfiguration(fileName);
            config.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(config, fileName, span);
            if (!sourceFile) {
                return [];
            }
            const tree = config.getService().getNavigationTree(fileName);
            return Array.from(this._flattenNavigationTreeItem(tree, null, sourceFile));
        });
    }
    /**
     * The workspace references request is sent from the client to the server to locate project-wide
     * references to a symbol given its description / metadata.
     */
    workspaceXreferences(params, span = new opentracing_1.Span()) {
        return rxjs_1.Observable.from(this.projectManager.ensureAllFiles(span))
            .mergeMap(() => this.projectManager.configurations())
            .filter(config => !(params.hints && params.hints.dependeePackageName && params.hints.dependeePackageName !== config.getPackageName()))
            .mergeMap(config => {
            config.ensureAllFiles(span);
            return rxjs_1.Observable.from(config.getService().getProgram().getSourceFiles())
                .filter(source => !util.toUnixPath(source.fileName).includes('/node_modules/'))
                .mergeMap(source => 
            // Iterate AST of source file
            rxjs_1.Observable.from(ast_1.walkMostAST(source))
                .filter((node) => node.kind === ts.SyntaxKind.Identifier)
                .mergeMap(node => 
            // Get DefinitionInformations at the node
            rxjs_1.Observable.from(config.getService().getDefinitionAtPosition(source.fileName, node.pos + 1) || [])
                .map(definition => util.defInfoToSymbolDescriptor(definition))
                .filter(symbol => util.symbolDescriptorMatch(params.query, symbol))
                .map(symbol => ({
                symbol,
                reference: {
                    uri: this._defUri(source.fileName),
                    range: {
                        start: ts.getLineAndCharacterOfPosition(source, node.pos),
                        end: ts.getLineAndCharacterOfPosition(source, node.end)
                    }
                }
            }))));
        })
            .toArray();
    }
    /**
     * This method returns metadata about the package(s) defined in a workspace and a list of
     * dependencies for each package.
     *
     * This method is necessary to implement cross-repository jump-to-def when it is not possible to
     * resolve the global location of the definition from data present or derived from the local
     * workspace. For example, a package manager might not include information about the source
     * repository of each dependency. In this case, definition resolution requires mapping from
     * package descriptor to repository revision URL. A reverse index can be constructed from calls
     * to workspace/xpackages to provide an efficient mapping.
     */
    workspaceXpackages(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure(span))
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => rxjs_1.Observable.from(this.updater.ensure(uri))
            .map(() => JSON.parse(this.inMemoryFileSystem.getContent(uri))))
            .mergeMap(packageJson => {
            const packageDescriptor = {
                name: packageJson.name,
                version: packageJson.version,
                repoURL: packageJson.repository && packageJson.repository.url || undefined
            };
            // Collect all dependencies for this package.json
            return rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
                .filter(key => packageJson[key])
                .mergeMap(key => lodash_1.toPairs(packageJson[key]))
                .map(([name, version]) => ({
                attributes: {
                    name,
                    version
                },
                hints: {
                    dependeePackageName: packageJson.name
                }
            }))
                .toArray()
                .map(dependencies => ({
                package: packageDescriptor,
                dependencies
            }));
        })
            .toArray();
    }
    /**
     * Returns all dependencies of a workspace.
     * Superseded by workspace/xpackages
     */
    workspaceXdependencies(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure())
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => rxjs_1.Observable.from(this.updater.ensure(uri))
            .map(() => JSON.parse(this.inMemoryFileSystem.getContent(uri))))
            .mergeMap(packageJson => rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
            .filter(key => packageJson[key])
            .mergeMap(key => lodash_1.toPairs(packageJson[key]))
            .map(([name, version]) => ({
            attributes: {
                name,
                version
            },
            hints: {
                dependeePackageName: packageJson.name
            }
        })))
            .toArray();
    }
    /**
     * The Completion request is sent from the client to the server to compute completion items at a
     * given cursor position. Completion items are presented in the
     * [IntelliSense](https://code.visualstudio.com/docs/editor/editingevolved#_intellisense) user
     * interface. If computing full completion items is expensive, servers can additionally provide
     * a handler for the completion item resolve request ('completionItem/resolve'). This request is
     * sent when a completion item is selected in the user interface. A typically use case is for
     * example: the 'textDocument/completion' request doesn't fill in the `documentation` property
     * for returned completion items since it is expensive to compute. When the item is selected in
     * the user interface then a 'completionItem/resolve' request is sent with the selected
     * completion item as a param. The returned completion item should have the documentation
     * property filled in.
     */
    textDocumentCompletion(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                return vscode_languageserver_1.CompletionList.create();
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const completions = configuration.getService().getCompletionsAtPosition(fileName, offset);
            if (completions == null) {
                return vscode_languageserver_1.CompletionList.create();
            }
            return vscode_languageserver_1.CompletionList.create(completions.entries.map(item => {
                const ret = vscode_languageserver_1.CompletionItem.create(item.name);
                const kind = completionKinds[item.kind];
                if (kind) {
                    ret.kind = kind;
                }
                if (item.sortText) {
                    ret.sortText = item.sortText;
                }
                const details = configuration.getService().getCompletionEntryDetails(fileName, offset, item.name);
                if (details) {
                    ret.documentation = ts.displayPartsToString(details.documentation);
                    ret.detail = ts.displayPartsToString(details.displayParts);
                }
                return ret;
            }));
        });
    }
    /**
     * The signature help request is sent from the client to the server to request signature
     * information at a given cursor position.
     */
    textDocumentSignatureHelp(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve signature are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const filePath = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(filePath);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, filePath, span);
            if (!sourceFile) {
                throw new Error(`expected source file ${filePath} to exist in configuration`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const signatures = configuration.getService().getSignatureHelpItems(filePath, offset);
            if (!signatures) {
                return { signatures: [], activeParameter: 0, activeSignature: 0 };
            }
            const signatureInformations = signatures.items.map(item => {
                const prefix = ts.displayPartsToString(item.prefixDisplayParts);
                const params = item.parameters.map(p => ts.displayPartsToString(p.displayParts)).join(', ');
                const suffix = ts.displayPartsToString(item.suffixDisplayParts);
                const parameters = item.parameters.map(p => {
                    return vscode_languageserver_1.ParameterInformation.create(ts.displayPartsToString(p.displayParts), ts.displayPartsToString(p.documentation));
                });
                return vscode_languageserver_1.SignatureInformation.create(prefix + params + suffix, ts.displayPartsToString(item.documentation), ...parameters);
            });
            return {
                signatures: signatureInformations,
                activeSignature: signatures.selectedItemIndex,
                activeParameter: signatures.argumentIndex
            };
        });
    }
    /**
     * The document open notification is sent from the client to the server to signal newly opened
     * text documents. The document's truth is now managed by the client and the server must not try
     * to read the document's truth using the document's uri.
     */
    textDocumentDidOpen(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            // Ensure files needed for most operations are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            this.projectManager.didOpen(util.uri2path(uri), params.textDocument.text);
        });
    }
    /**
     * The document change notification is sent from the client to the server to signal changes to a
     * text document. In 2.0 the shape of the params has changed to include proper version numbers
     * and language ids.
     */
    textDocumentDidChange(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            let text = null;
            params.contentChanges.forEach(change => {
                if (change.range || change.rangeLength) {
                    throw new Error('incremental updates in textDocument/didChange not supported for file ' + params.textDocument.uri);
                }
                text = change.text;
            });
            if (!text) {
                return;
            }
            this.projectManager.didChange(util.uri2path(uri), text);
        });
    }
    /**
     * The document save notification is sent from the client to the server when the document was
     * saved in the client.
     */
    textDocumentDidSave(params) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            // TODO don't use "relative" URI
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            this.projectManager.didSave(util.uri2path(uri));
        });
    }
    /**
     * The document close notification is sent from the client to the server when the document got
     * closed in the client. The document's truth now exists where the document's uri points to
     * (e.g. if the document's uri is a file uri the truth now exists on disk).
     */
    textDocumentDidClose(params) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            // TODO don't use "relative" URI
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            this.projectManager.didClose(util.uri2path(uri));
        });
    }
    /**
     * Fetches (or creates if needed) source file object for a given file name
     *
     * @param configuration project configuration
     * @param fileName file name to fetch source file for or create it
     * @param span Span for tracing
     */
    _getSourceFile(configuration, fileName, span = new opentracing_1.Span()) {
        let program = configuration.getProgram();
        if (!program) {
            return undefined;
        }
        const sourceFile = program.getSourceFile(fileName);
        if (sourceFile) {
            return sourceFile;
        }
        if (!this.projectManager.hasFile(fileName)) {
            return undefined;
        }
        configuration.getHost().addFile(fileName);
        // Update program object
        configuration.syncProgram(span);
        program = configuration.getProgram();
        return program && program.getSourceFile(fileName);
    }
    /**
     * transformNavItem transforms a NavigateToItem instance to a SymbolInformation instance
     */
    _transformNavItem(program, item) {
        const sourceFile = program.getSourceFile(item.fileName);
        if (!sourceFile) {
            throw new Error('source file "' + item.fileName + '" does not exist');
        }
        const start = ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start);
        const end = ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start + item.textSpan.length);
        return vscode_languageserver_1.SymbolInformation.create(item.name, util.convertStringtoSymbolKind(item.kind), vscode_languageserver_1.Range.create(start.line, start.character, end.line, end.character), this._defUri(item.fileName), item.containerName);
    }
    /**
     * Returns an Iterator for all symbols in a given config
     *
     * Note: This method is not traced because it returns an Iterator that may produce values lazily
     *
     * @param config The ProjectConfiguration to search
     * @param query A text or SymbolDescriptor query
     * @param limit An optional limit that is passed to TypeScript
     * @return Iterator that emits SymbolInformations
     */
    _collectWorkspaceSymbols(config, query, limit = Infinity, span = new opentracing_1.Span()) {
        config.ensureAllFiles(span);
        const program = config.getProgram();
        if (!program) {
            return iterare_1.default([]);
        }
        if (query) {
            let items;
            if (typeof query === 'string') {
                // Query by text query
                items = config.getService().getNavigateToItems(query, limit, undefined, false);
            }
            else {
                // Query by name
                const packageName = config.getPackageName();
                const packageDescriptor = packageName && { name: packageName } || undefined;
                items = iterare_1.default(config.getService().getNavigateToItems(query.name || '', limit, undefined, false))
                    .filter(item => util.symbolDescriptorMatch(query, {
                    kind: item.kind,
                    name: item.name,
                    containerKind: item.containerKind,
                    containerName: item.containerName,
                    package: packageDescriptor
                }));
            }
            return iterare_1.default(items)
                .map(item => this._transformNavItem(program, item))
                .filter(symbolInformation => util.isLocalUri(symbolInformation.location.uri));
        }
        else {
            // An empty query uses a different algorithm to iterate all files and aggregate the symbols per-file to get all symbols
            // TODO make all implementations use this? It has the advantage of being streamable and cancellable
            return iterare_1.default(this._getNavigationTreeItems(config)).take(limit);
        }
    }
    /**
     * Transforms definition's file name to URI. If definition belongs to TypeScript library,
     * returns git://github.com/Microsoft/TypeScript URL, otherwise returns file:// one
     */
    _defUri(filePath) {
        if (memfs_1.isTypeScriptLibrary(filePath)) {
            return 'git://github.com/Microsoft/TypeScript?v' + ts.version + '#lib/' + path_.basename(filePath);
        }
        return util.path2uri(this.root, filePath);
    }
    /**
     * Fetches up to limit navigation bar items from given project, flattens them
     */
    _getNavigationTreeItems(configuration) {
        const program = configuration.getProgram();
        if (!program) {
            return iterare_1.default([]);
        }
        return iterare_1.default(program.getSourceFiles())
            .filter(sourceFile => !memfs_1.isTypeScriptLibrary(sourceFile.fileName))
            .map(sourceFile => {
            try {
                const tree = configuration.getService().getNavigationTree(sourceFile.fileName);
                return this._flattenNavigationTreeItem(tree, null, sourceFile);
            }
            catch (e) {
                this.logger.error('Could not get navigation tree for file', sourceFile.fileName);
                return [];
            }
        })
            .flatten();
    }
    /**
     * Flattens navigation tree by emitting acceptable NavigationTreeItems as SymbolInformations.
     * Some items (source files, modules) may be excluded
     */
    *_flattenNavigationTreeItem(item, parent, sourceFile) {
        const acceptable = TypeScriptService.isAcceptableNavigationTreeItem(item);
        if (acceptable) {
            yield this._transformNavigationTreeItem(item, parent, sourceFile);
        }
        if (item.childItems) {
            for (const childItem of item.childItems) {
                yield* this._flattenNavigationTreeItem(childItem, acceptable ? item : null, sourceFile);
            }
        }
    }
    /**
     * Transforms NavigationTree to SymbolInformation
     */
    _transformNavigationTreeItem(item, parent, sourceFile) {
        const span = item.spans[0];
        const start = ts.getLineAndCharacterOfPosition(sourceFile, span.start);
        const end = ts.getLineAndCharacterOfPosition(sourceFile, span.start + span.length);
        return {
            name: item.text,
            kind: util.convertStringtoSymbolKind(item.kind),
            location: {
                uri: this._defUri(sourceFile.fileName),
                range: { start, end }
            },
            containerName: parent ? parent.text : ''
        };
    }
    /**
     * @return true if navigation tree item is acceptable for inclusion into workspace/symbols
     */
    static isAcceptableNavigationTreeItem(item) {
        // modules and source files should be excluded
        if ([ts.ScriptElementKind.moduleElement, 'sourcefile'].indexOf(item.kind) >= 0) {
            return false;
        }
        // special items may start with ", (, [, or <
        if (/^[<\(\[\"]/.test(item.text)) {
            return false;
        }
        // magic words
        if (['default', 'constructor', 'new()'].indexOf(item.text) >= 0) {
            return false;
        }
        return true;
    }
}
exports.TypeScriptService = TypeScriptService;
/**
 * Maps string-based CompletionEntry::kind to enum-based CompletionItemKind
 */
const completionKinds = {
    class: vscode_languageserver_1.CompletionItemKind.Class,
    constructor: vscode_languageserver_1.CompletionItemKind.Constructor,
    enum: vscode_languageserver_1.CompletionItemKind.Enum,
    field: vscode_languageserver_1.CompletionItemKind.Field,
    file: vscode_languageserver_1.CompletionItemKind.File,
    function: vscode_languageserver_1.CompletionItemKind.Function,
    interface: vscode_languageserver_1.CompletionItemKind.Interface,
    keyword: vscode_languageserver_1.CompletionItemKind.Keyword,
    method: vscode_languageserver_1.CompletionItemKind.Method,
    module: vscode_languageserver_1.CompletionItemKind.Module,
    property: vscode_languageserver_1.CompletionItemKind.Property,
    reference: vscode_languageserver_1.CompletionItemKind.Reference,
    snippet: vscode_languageserver_1.CompletionItemKind.Snippet,
    text: vscode_languageserver_1.CompletionItemKind.Text,
    unit: vscode_languageserver_1.CompletionItemKind.Unit,
    value: vscode_languageserver_1.CompletionItemKind.Value,
    variable: vscode_languageserver_1.CompletionItemKind.Variable
};
//# sourceMappingURL=typescript-service.js.map